import openai
import os
from autogpt.config import Config
from typing import Optional

def setup_config(
        azure_config_file:Optional[str] = "",
        debug:bool = False,
) -> Config:
    CFG = Config()
    if azure_config_file != "":
        CFG.set_openai_api_key(os.getenv("AZURE_OPENAI_API_KEY", ""))
        CFG.use_azure=True
        CFG.load_azure_config(config_file=azure_config_file)

    CFG.set_debug_mode(debug)

    openai.api_type = CFG.openai_api_type
    openai.api_base = CFG.openai_api_base
    openai.api_version = CFG.openai_api_version
    openai.api_key = CFG.openai_api_key
    return CFG


from typing import Dict, List, Dict
from autogpt.llm_utils import create_chat_completion

def create_message(prompt: str) -> List[Dict]:
    """Create a message for the chat completion

    Args:
        chunk (str): The chunk of text to summarize
        question (str): The question to answer

    Returns:
        Dict[str, str]: The message to send to the chat completion
    """
    return [
        {
            "role": "system",
            "content": "You are an AI assistant."
        },
        {
            "role": "user",
            "content": prompt
        }
    ]

class llm_uils:

    model: str

    def __init__(self, model: str = "gpt-3.5-turbo"):
        self.model = model

    def text_completion(self, prompt: str) -> str:
        messages = create_message(prompt)
        response = create_chat_completion(
            model=self.model,
            messages=messages,
        )

        return response


import json, time
from typing import Dict, List, Dict, Optional
import logging

from autogpt.llm_utils import create_chat_completion
from autogpt.config import Config

from autoxx.tools.web_search.search import web_search
from autoxx.tools.llm.utils import llm_uils, create_message


class AGIExecutor:
    def __init__(self, objective:str, max_tasks_count: Optional[int] = 5):
        if objective is None or len(objective) <= 1:  
            raise ValueError("Must provide an objective")  

        self.objective = objective
        self.max_tasks_count = max_tasks_count

        self.task_list = self.init_task_list(objective)
        self.status = "initizing"
        self.error_message = None
        self.final_answer = None

    def get_task_by_id(self, task_id: int):
        for task in self.task_list:
            if task["id"] == task_id:
                return task
        return None

    def init_task_list(self, objective: str) -> List[Dict]:
        task_id_counter = 0
        prompt = (
            f"You are a QA AI tasked with creating tasks for the following objective: {objective}.\n"
            f"Create new tasks based on the following plan delimited by triple backtick if necessary for the objective. Limit tasks types to those that can be completed with the available tools listed below. Task description should be detailed. "
            f"The plan:```decelop a task list```.\n"
            f"The current tool option is [text-completion, web-search] only. What each tool does is as follows:\n"
            f"web-search: It supports only searching for information from the Internet. For tasks using [web-search], provide the search query, and only the search query to use (eg. not 'research waterproof shoes, but 'waterproof shoes').\n"
            f"text-completion: it is a text completion tool that can be used for content extraction, summarization, copywrite, translation, etc., and can also be used for LLM-based QA\n"
            f"Make sure all task IDs are in chronological order.\n"
            f"For efficiency, let's think step by step and create the tasks (up to 3 tasks) that are most critical to objective.\n"
            f"An example of the desired output format is:"
            "[{\"id\": 1, \"task\": \"https://untapped.vc\", \"tool\": \"web-scrape\", \"dependent_task_ids\": [], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}, {\"id\": 2, \"task\": \"Consider additional insights that can be reasoned from the results of...\", \"tool\": \"text-completion\", \"dependent_task_ids\": [1], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}, {\"id\": 3, \"task\": \"Untapped Capital\", \"tool\": \"web-search\", \"dependent_task_ids\": [], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}].\n"
            f"Ensure the response can be parsed by Python json.loads. JSON TASK LIST="
        )

        messages = create_message(prompt)
        response = create_chat_completion(
            model=Config().fast_llm_model,
            messages=messages,
        )

        new_tasks = json.loads(response)
        task_list = []
        for new_task in new_tasks:
            task_id_counter += 1
            new_task.update({"id": task_id_counter})
            task_list.append(new_task)
        return task_list

    def orchestrate_task(self, last_task: Dict) -> List[Dict]:
        original_task_list = self.task_list.copy()
        minified_task_list = [{k: v for k, v in task.items() if k != "result"} for task in self.task_list]
        result = last_task['result'][0:4000] #come up with better solution later.
        
        prompt = (
            f"You are a QA AI assistant tasked with cleaning the formatting of and reprioritizing the following tasks delimited by triple backtick:```{minified_task_list}```\n"
            f"the last completed task (task id = {last_task['id']}, tool = {last_task['tool']}) has the following result delimited by triple backtick: ```{last_task['result']}```\n\n"
            f"Your task: Consider the ultimate objective of your team is to answer the question {self.objective} and reflect on the result of task that has been complted, re-create the task list (up to 5 tasks) to answer the question.\n" 
            f"The current tool option is [text-completion, web-search] only. What each tool does is as follows:\n"
            f"web-search: It supports searching for information on the Internet. For tasks using [web-search], provide the search query, and only the search query to use (eg. not 'research waterproof shoes, but 'waterproof shoes').\n"
            f"text-completion: IT is a tool that can be used for content extraction, summarization, copywrite, translation, etc., and can also be used for LLM-based QA\n"
            f"dependent_task_ids should always be an empty array, or an array of numbers representing the task ID it should pull results from."
            f"Make sure all task IDs are in chronological order.\n"
            f"Don not remove the completed tasks.\n"
            f"The last task is always to provide a final answer for customer.\n"
            f"An example of the desired output format is:"
            "[{\"id\": 1, \"task\": \"https://untapped.vc\", \"tool\": \"web-scrape\", \"dependent_task_ids\": [], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}, {\"id\": 2, \"task\": \"Consider additional insights that can be reasoned from the results of...\", \"tool\": \"text-completion\", \"dependent_task_ids\": [1], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}, {\"id\": 3, \"task\": \"Untapped Capital\", \"tool\": \"web-search\", \"dependent_task_ids\": [], \"status\": \"incomplete\", \"result\": null, \"result_summary\": null}].\n"
            f"Ensure the response can be parsed by Python json.loads. JSON TASK LIST="
        )

        messages = create_message(prompt)
        result = create_chat_completion(
            model=Config().fast_llm_model,
            messages=messages,
        )

        try:
            tmp_task_list = json.loads(result)
        except Exception as error:
            print(f"fail to decode ({error}): {result}")
            return original_task_list
        # Add the 'result' field back in

        for updated_task, original_task in zip(tmp_task_list, original_task_list):
            if "result" in original_task:
                updated_task["result"] = original_task["result"]

        return tmp_task_list

    ### Agent functions ##############################
    def execute_task(self, task: Dict) -> Dict:
        # Check if dependent_task_id is completed
        dependent_task_ids = []
        dependent_task_prompt = ""
        if task["dependent_task_ids"]:
            if isinstance(task["dependent_task_ids"], list):
                dependent_task_ids = task["dependent_task_ids"]
            else:
                dependent_task_ids = [task["dependent_task_ids"]]

            for dependent_task_id in dependent_task_ids:
                dependent_task = self.get_task_by_id(dependent_task_id)
                if dependent_task and dependent_task["status"] != "completed":
                    return
                elif dependent_task:
                    dependent_task_prompt += f"\ntask ({dependent_task['id']}. {dependent_task['task']}) result: {dependent_task['result']}"

        # Execute task
        
        task_prompt = f"Complete your assigned task based on the objective: {self.objective}. Your task: {task['task']}"
        if dependent_task_prompt != "":
            task_prompt += f"\nThe previous tasks: {dependent_task_prompt}"

        task_prompt += "\nResponse:"
        if task["tool"] == "text-completion":
            result = llm_uils().text_completion(task_prompt)
            summary_result_prompt = f"Please summarize the following text:\n{result}\nSummary:"
            task["result_summary"] = llm_uils().text_completion(summary_result_prompt)
        elif task["tool"] == "web-search":
            result = str(web_search(str(task['task'])))
            summary_result_prompt = f"Please summarize the following text:\n{result}\nSummary:"
            task["result_summary"] = llm_uils().text_completion(summary_result_prompt)
        elif task["tool"] == "search-IT-knowledge-base":
            result = str(self.it_kb.similarity_search(str(task['task'])))
            summary_result_prompt = f"Please summarize the following text:\n{result}\nSummary:"
            task["result_summary"] = llm_uils().text_completion(summary_result_prompt)
        else:
            result = "Unknown tool"
        
        # Update task status and result
        task["status"] = "completed"
        task["result"] = result
        return task
    
    def execute(self) -> str:
        self.status = "running"
        # Main loop
        while any(task["status"] == "incomplete" for task in self.task_list):
              # Filter out incomplete tasks
            incomplete_tasks = [task for task in self.task_list if task["status"] == "incomplete"]
            logging.info(f"\033[95m\033[1m INCOMPLETE TASK LIST({self.objective}) \033[0m\033[0m {[str(t['id'])+': '+t['task'] + '[' + t['tool'] + ']' for t in incomplete_tasks]}")
            if incomplete_tasks:
                # Sort tasks by ID
                incomplete_tasks.sort(key=lambda x: x["id"])

                # Pull the first task
                task = incomplete_tasks[0]
                logging.info(f"\033[92m\033[1m NEXT TASK({self.objective}) \033[0m\033[0m {task['id']}:{task['task']}[{task['tool']}]")

                # Execute task & call task manager from function
                try:
                    completed_task = self.execute_task(task)
                    logging.info(f"\033[93m\033[1m TASK RESULT({self.objective}) \033[0m\033[0m {completed_task['result']}")
                except Exception as e:  
                    logging.error(f"An error occurred while executing the task({self.objective}): {e.with_traceback()}")
                    self.status = "error"
                    self.error_message = str(e)  
                    return str(e)
                if len(self.task_list) - len(incomplete_tasks) + 1 < self.max_tasks_count or len(incomplete_tasks) > 1:
                    logging.info(f"\033[90m\033[3m" + "\nOrchestrate_task list({self.objective})...\n" + "\033[0m")
                    self.task_list = self.orchestrate_task(completed_task)

            time.sleep(1)  # Sleep before checking the task list again
        self.status = "completed"

    def stat(self) -> Dict:
        return {"task_list": self.task_list, "status": self.status, "error_message": self.error_message}


import logging, time, threading
from typing import Dict, Optional
from autoxx.agent.babyagi.agi import AGIExecutor
  
class TaskManager:  
    def __init__(self):  
        self.tasks = {}  
  
    def create_task(self, objective: str, max_tasks_count: Optional[int] = 3):  
        if objective in self.tasks:
            if self.tasks[objective].status == "error":
                task = self.tasks[objective]
                task_thread = threading.Thread(target=task.execute)  
                task_thread.start()
                return
            else:
                raise ValueError(f"Task found with objective: {objective}")
        
        task = AGIExecutor(objective, max_tasks_count) 
        self.tasks[objective] = task  
  
        # Start task execution in a separate thread  
        task_thread = threading.Thread(target=task.execute)  
        task_thread.start()
  
    def get_task_stat(self, objective: str) -> Dict:  
        if objective not in self.tasks:  
            raise ValueError(f"No task found with objective: {objective}")  
  
        return self.tasks[objective].stat()


import pinecone
import os
from langchain.chat_models import ChatOpenAI
from autogpt.config import Config
from llama_index.embeddings.openai import OpenAIEmbedding

from llama_index import (LLMPredictor,
                        ServiceContext,
                        StorageContext)
from llama_index.indices.loading import load_index_from_storage
from llama_index.storage.docstore import MongoDocumentStore
from llama_index.storage.index_store import MongoIndexStore
from llama_index.vector_stores import PineconeVectorStore
from llama_index.readers.schema.base import Document
from typing import List, Dict
from llama_index.node_parser import SimpleNodeParser
from llama_index.langchain_helpers.text_splitter import TokenTextSplitter
from llama_index.callbacks.base import CallbackManager

class knowleage_bot:

    def __init__(self, corpus:str, model:str = "gpt-3.5-turbo"):
        self.corpus = corpus
        self.model = model
        mongodb_dbname = self.corpus

        mongodb_password = os.getenv("MONGODB_PASSWORD")
        assert mongodb_password is not None, "Please set MONGODB_PASSWORD environment variable"
        mongodb_user = os.getenv("MONGODB_USER") or "raga"
        mongodb_url = f"mongodb+srv://{mongodb_user}:{mongodb_password}@cluster0.spj0g.mongodb.net/?retryWrites=true&w=majority"

        # create mongodb docstore, indexstore
        self.docstore = MongoDocumentStore.from_uri(uri=mongodb_url, db_name=mongodb_dbname)
        self.index_store = MongoIndexStore.from_uri(uri=mongodb_url, db_name=mongodb_dbname)
        pinecone_api_key = os.getenv("PINECONE_API_KEY")
        assert pinecone_api_key is not None, "Please set PINECONE_API_KEY environment variable"
        pinecone_environment =  os.getenv("PINECONE_ENVIRONMENT") or "us-central1-gcp"
        pinecone_index_name = f"{self.corpus}-index"

        pinecone.init(api_key=pinecone_api_key, environment=pinecone_environment)
        pinecone_index = pinecone.Index(pinecone_index_name)
        self.vector_store = PineconeVectorStore(pinecone_index=pinecone_index)

        storage_context = StorageContext.from_defaults(
            docstore=self.docstore,
            index_store=self.index_store,
            vector_store=self.vector_store,
        )

        config = Config()
        if config.use_azure:
            llm = ChatOpenAI(model_name= config.fast_llm_model, model_kwargs={
                "api_key": config.openai_api_key,
                "api_base": config.openai_api_base,
                "api_type": config.openai_api_type,
                "api_version": config.openai_api_version,
                "deployment_id": config.get_azure_deployment_id_for_model(config.fast_llm_model),
            })
            llm_predictor_chatgpt = LLMPredictor(
                llm=llm
            )
            embed_model = OpenAIEmbedding(
                deployment_name=config.get_azure_deployment_id_for_model("text-embedding-ada-002"),
                api_key=config.openai_api_key,
                api_base=config.openai_api_base,
                api_type=config.openai_api_type,
                api_version=config.openai_api_version,
            )
        else:
            print(config.use_azure, config.openai_api_key)
            llm_predictor_chatgpt = LLMPredictor(
                llm=ChatOpenAI(temperature=0, model_name="gpt-3.5-turbo")
            )
            embed_model = OpenAIEmbedding()
        service_context = ServiceContext.from_defaults(
            llm_predictor=llm_predictor_chatgpt, chunk_size_limit=1024, embed_model=embed_model,
            node_parser=SimpleNodeParser(
                text_splitter=TokenTextSplitter(
                    callback_manager=CallbackManager([]),
                    chunk_size = 2048,
            ))
        )
        self.index = load_index_from_storage(
            storage_context=storage_context, service_context=service_context
        )

    def query(self, query:str):
        try:
            engine = self.index.as_query_engine()
            return engine.query(query)
        except Exception as e:
            print("Error:", e)
            raise Exception(f"failed to finish query: {str(e)}")

    def similarity_search(self, query:str):
        try:
            retriever = self.index.as_retriever()
            nodes =  retriever.retrieve(query)
        except Exception as e:
            print("Error:", e)
            raise Exception(f"failed to retrieve nodes: {str(e)}")

        return nodes

    def upsert_document(self, documents: List[Document]) -> None:
        self.index.docstore.add_documents(documents, allow_update=True)
        for document in documents:
            nodes = self.index.service_context.node_parser.get_nodes_from_documents([document])
            self.index.insert_nodes(nodes)

    def delete_document(self, document_id: str) -> None:
        self.index.delete(document_id)
        self.index.docstore.delete_document(document_id)

    def retrieve_document(self, document_ids: List[str]) ->  List[Document]:
        if document_ids is None or len(document_ids) == 0:
            return [doc for _, doc in self.index.docstore.docs.items()]

        documents = []
        for document_id in document_ids:
            document = self.index.docstore.get_document(document_id)
            documents.append(document)
        return documents
